@@ -34,8 +34,12 @@ The main interface is :class:`TrinoQuery`: ::
 """
 
 import copy
+import functools
 import os
+import random
 import re
+import threading
+import time
 import urllib.parse
 from datetime import datetime, timedelta, timezone
 from decimal import Decimal
@@ -46,9 +50,8 @@ import requests
 
 import trino.logging
 from trino import constants, exceptions
-from trino.transaction import NO_TRANSACTION
 
-__all__ = ["TrinoQuery", "TrinoRequest", "PROXIES"]
+__all__ = ["ClientSession", "TrinoQuery", "TrinoRequest", "PROXIES"]
 
 logger = trino.logging.get_logger(__name__)
 
@@ -67,38 +70,142 @@ NAN = float("nan")
 
 
 class ClientSession(object):
+    """
+    Manage the current Client Session properties of a specific connection. This class is thread-safe.
+
+    :param user: associated with the query. It is useful for access control
+                 and query scheduling.
+    :param source: associated with the query. It is useful for access
+                   control and query scheduling.
+    :param catalog: to query. The *catalog* is associated with a Trino
+                    connector. This variable sets the default catalog used
+                    by SQL statements. For example, if *catalog* is set
+                    to ``some_catalog``, the SQL statement
+                    ``SELECT * FROM some_schema.some_table`` will actually
+                    query the table
+                    ``some_catalog.some_schema.some_table``.
+    :param schema: to query. The *schema* is a logical abstraction to group
+                   table. This variable sets the default schema used by
+                   SQL statements. For example, if *schema* is set to
+                   ``some_schema``, the SQL statement
+                   ``SELECT * FROM some_table`` will actually query the
+                   table ``some_catalog.some_schema.some_table``.
+    :param properties: set specific Trino behavior for the current
+                               session. Please refer to the output of
+                               ``SHOW SESSION`` to check the available
+                               properties.
+    :param headers: HTTP headers to POST/GET in the HTTP requests
+    :param extra_credential: extra credentials. as list of ``(key, value)``
+                             tuples.
+    :param client_tags: Client tags as list of strings.
+    :param role: role for the current session. Some connectors do not
+                 support role management. See connector documentation for more details.
+    """
+
     def __init__(
         self,
-        catalog,
-        schema,
-        source,
-        user,
-        properties=None,
-        headers=None,
-        transaction_id=None,
-        extra_credential=None,
-        client_tags=None
+        user: str,
+        catalog: str = None,
+        schema: str = None,
+        source: str = None,
+        properties: Dict[str, str] = None,
+        headers: Dict[str, str] = None,
+        transaction_id: str = None,
+        extra_credential: List[Tuple[str, str]] = None,
+        client_tags: List[str] = None,
+        role: str = None,
     ):
-        self.catalog = catalog
-        self.schema = schema
-        self.source = source
-        self.user = user
-        if properties is None:
-            properties = {}
-        self._properties = properties
+        self._user = user
+        self._catalog = catalog
+        self._schema = schema
+        self._source = source
+        self._properties = properties or {}
         self._headers = headers or {}
-        self.transaction_id = transaction_id
-        self.extra_credential = extra_credential
-        self.client_tags = client_tags
+        self._transaction_id = transaction_id
+        self._extra_credential = extra_credential
+        self._client_tags = client_tags
+        self._role = role
+        self._object_lock = threading.Lock()
+
+    @property
+    def user(self):
+        return self._user
+
+    @property
+    def catalog(self):
+        with self._object_lock:
+            return self._catalog
+
+    @catalog.setter
+    def catalog(self, catalog):
+        with self._object_lock:
+            self._catalog = catalog
+
+    @property
+    def schema(self):
+        with self._object_lock:
+            return self._schema
+
+    @schema.setter
+    def schema(self, schema):
+        with self._object_lock:
+            self._schema = schema
+
+    @property
+    def source(self):
+        return self._source
 
     @property
     def properties(self):
-        return self._properties
+        with self._object_lock:
+            return self._properties
+
+    @properties.setter
+    def properties(self, properties):
+        with self._object_lock:
+            self._properties = properties
 
     @property
     def headers(self):
         return self._headers
 
+    @property
+    def transaction_id(self):
+        with self._object_lock:
+            return self._transaction_id
+
+    @transaction_id.setter
+    def transaction_id(self, transaction_id):
+        with self._object_lock:
+            self._transaction_id = transaction_id
+
+    @property
+    def extra_credential(self):
+        return self._extra_credential
+
+    @property
+    def client_tags(self):
+        return self._client_tags
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        del state["_object_lock"]
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        self._object_lock = threading.Lock()
+
+    @property
+    def role(self):
+        with self._object_lock:
+            return self._role
+
+    @role.setter
+    def role(self, role):
+        with self._object_lock:
+            self._role = role
+
 
 def get_header_values(headers, header):
     return [val.strip() for val in headers[header].split(",")]
@@ -137,39 +244,43 @@ class TrinoStatus(object):
         )
 
 
+class _DelayExponential(object):
+    def __init__(
+            self, base=0.1, exponent=2, jitter=True, max_delay=2 * 3600  # 100ms  # 2 hours
+    ):
+        self._base = base
+        self._exponent = exponent
+        self._jitter = jitter
+        self._max_delay = max_delay
+
+    def __call__(self, attempt):
+        delay = float(self._base) * (self._exponent ** attempt)
+        if self._jitter:
+            delay *= random.random()
+        delay = min(float(self._max_delay), delay)
+        return delay
+
+
+class _RetryWithExponentialBackoff(object):
+    def __init__(
+            self, base=0.1, exponent=2, jitter=True, max_delay=2 * 3600  # 100ms  # 2 hours
+    ):
+        self._get_delay = _DelayExponential(base, exponent, jitter, max_delay)
+
+    def retry(self, func, args, kwargs, err, attempt):
+        delay = self._get_delay(attempt)
+        time.sleep(delay)
+
+
 class TrinoRequest(object):
     """
     Manage the HTTP requests of a Trino query.
 
     :param host: name of the coordinator
     :param port: TCP port to connect to the coordinator
-    :param user: associated with the query. It is useful for access control
-                 and query scheduling.
-    :param source: associated with the query. It is useful for access
-                   control and query scheduling.
-    :param catalog: to query. The *catalog* is associated with a Trino
-                    connector. This variable sets the default catalog used
-                    by SQL statements. For example, if *catalog* is set
-                    to ``some_catalog``, the SQL statement
-                    ``SELECT * FROM some_schema.some_table`` will actually
-                    query the table
-                    ``some_catalog.some_schema.some_table``.
-    :param schema: to query. The *schema* is a logical abstraction to group
-                   table. This variable sets the default schema used by
-                   SQL statements. For eample, if *schema* is set to
-                   ``some_schema``, the SQL statement
-                   ``SELECT * FROM some_table`` will actually query the
-                   table ``some_catalog.some_schema.some_table``.
-    :param session_properties: set specific Trino behavior for the current
-                               session. Please refer to the output of
-                               ``SHOW SESSION`` to check the available
-                               properties.
-    :param http_headers: HTTP headers to post/get in the HTTP requests
     :param http_scheme: "http" or "https"
     :param auth: class that manages user authentication. ``None`` means no
                  authentication.
-    :param extra_credential: extra credentials. as list of ``(key, value)``
-                             tuples.
     :max_attempts: maximum number of attempts when sending HTTP requests. An
                    attempt is an HTTP request. 5 attempts means 4 retries.
     :request_timeout: How long (in seconds) to wait for the server to send
@@ -213,36 +324,17 @@ class TrinoRequest(object):
         self,
         host: str,
         port: int,
-        user: str,
-        source: str = None,
-        catalog: str = None,
-        schema: str = None,
-        session_properties: Optional[Dict[str, Any]] = None,
+        client_session: ClientSession,
         http_session: Any = None,
-        http_headers: Optional[Dict[str, str]] = None,
-        transaction_id: Optional[str] = NO_TRANSACTION,
         http_scheme: str = None,
         auth: Optional[Any] = constants.DEFAULT_AUTH,
-        extra_credential: Optional[List[Tuple[str, str]]] = None,
         redirect_handler: Any = None,
         max_attempts: int = MAX_ATTEMPTS,
         request_timeout: Union[float, Tuple[float, float]] = constants.DEFAULT_REQUEST_TIMEOUT,
-        handle_retry=exceptions.RetryWithExponentialBackoff(),
+        handle_retry=_RetryWithExponentialBackoff(),
         verify: bool = True,
-        client_tags: Optional[List[str]] = None
     ) -> None:
-        self._client_session = ClientSession(
-            catalog,
-            schema,
-            source,
-            user,
-            session_properties,
-            http_headers,
-            transaction_id,
-            extra_credential,
-            client_tags
-        )
-
+        self._client_session = client_session
         self._host = host
         self._port = port
         self._next_uri: Optional[str] = None
@@ -290,6 +382,7 @@ class TrinoRequest(object):
         headers[constants.HEADER_SCHEMA] = self._client_session.schema
         headers[constants.HEADER_SOURCE] = self._client_session.source
         headers[constants.HEADER_USER] = self._client_session.user
+        headers[constants.HEADER_ROLE] = self._client_session.role
         if self._client_session.client_tags is not None and len(self._client_session.client_tags) > 0:
             headers[constants.HEADER_CLIENT_TAGS] = ",".join(self._client_session.client_tags)
 
@@ -336,9 +429,9 @@ class TrinoRequest(object):
             self._delete = self._http_session.delete
             return
 
-        with_retry = exceptions.retry_with(
+        with_retry = _retry_with(
             self._handle_retry,
-            exceptions=self._exceptions,
+            handled_exceptions=self._exceptions,
             conditions=(
                 # need retry when there is no exception but the status code is 502, 503, or 504
                 lambda response: getattr(response, "status_code", None)
@@ -454,6 +547,15 @@ class TrinoRequest(object):
             ):
                 self._client_session.properties[key] = value
 
+        if constants.HEADER_SET_CATALOG in http_response.headers:
+            self._client_session.catalog = http_response.headers[constants.HEADER_SET_CATALOG]
+
+        if constants.HEADER_SET_SCHEMA in http_response.headers:
+            self._client_session.schema = http_response.headers[constants.HEADER_SET_SCHEMA]
+
+        if constants.HEADER_SET_ROLE in http_response.headers:
+            self._client_session.role = http_response.headers[constants.HEADER_SET_ROLE]
+
         self._next_uri = response.get("nextUri")
 
         return TrinoStatus(
@@ -504,7 +606,7 @@ class TrinoResult(object):
         # Initial fetch from the first POST request
         for row in self._rows:
             self._rownumber += 1
-            yield row
+            yield self._map_row(self._experimental_python_types, row, self._query.columns)
         self._rows = None
 
         # Subsequent fetches from GET requests until next_uri is empty.
@@ -513,15 +615,19 @@ class TrinoResult(object):
             for row in rows:
                 self._rownumber += 1
                 logger.debug("row %s", row)
-                if not self._experimental_python_types:
-                    yield row
-                else:
-                    yield self._map_to_python_types(row, self._query.columns)
+                yield self._map_row(self._experimental_python_types, row, self._query.columns)
 
     @property
     def response_headers(self):
         return self._query.response_headers
 
+    @classmethod
+    def _map_row(cls, experimental_python_types, row, columns):
+        if not experimental_python_types:
+            return row
+        else:
+            return cls._map_to_python_types(cls, row, columns)
+
     @classmethod
     def _map_to_python_type(cls, item: Tuple[Any, Dict]) -> Any:
         (value, data_type) = item
@@ -722,3 +828,32 @@ class TrinoQuery(object):
     @property
     def response_headers(self):
         return self._response_headers
+
+
+def _retry_with(handle_retry, handled_exceptions, conditions, max_attempts):
+    def wrapper(func):
+        @functools.wraps(func)
+        def decorated(*args, **kwargs):
+            error = None
+            result = None
+            for attempt in range(1, max_attempts + 1):
+                try:
+                    result = func(*args, **kwargs)
+                    if any(guard(result) for guard in conditions):
+                        handle_retry.retry(func, args, kwargs, None, attempt)
+                        continue
+                    return result
+                except Exception as err:
+                    error = err
+                    if any(isinstance(err, exc) for exc in handled_exceptions):
+                        handle_retry.retry(func, args, kwargs, err, attempt)
+                        continue
+                    break
+            logger.info("failed after %s attempts", attempt)
+            if error is not None:
+                raise error
+            return result
+
+        return decorated
+
+    return wrapper
