@@ -39,9 +39,15 @@ HEADER_SESSION = "X-Trino-Session"
 HEADER_SET_SESSION = "X-Trino-Set-Session"
 HEADER_CLEAR_SESSION = "X-Trino-Clear-Session"
 
+HEADER_ROLE = "X-Trino-Role"
+HEADER_SET_ROLE = "X-Trino-Set-Role"
+
 HEADER_STARTED_TRANSACTION = "X-Trino-Started-Transaction-Id"
 HEADER_TRANSACTION = "X-Trino-Transaction-Id"
 
 HEADER_PREPARED_STATEMENT = 'X-Trino-Prepared-Statement'
 HEADER_ADDED_PREPARE = 'X-Trino-Added-Prepare'
 HEADER_DEALLOCATED_PREPARE = 'X-Trino-Deallocated-Prepare'
+
+HEADER_SET_SCHEMA = "X-Trino-Set-Schema"
+HEADER_SET_CATALOG = "X-Trino-Set-Catalog"
